#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""09_collapse_map.py - An example of using toolkit.collapse_map in Gummworld2.

The results of this demo can be impressive on computers with fast CPUs.

Maps with many small tiles are great for designing maps with interesting detail.
But they take a lot longer for Python to process than larger tiles because there
are more objects in the working lists. toolkit.collapse_map() can help with
this.

The function takes a map with many small tiles and converts it to a map with
fewer, larger tiles for more efficient processing.

While running this demo use keys 0-9 to set the collapse level. 1 presents the
original map of 32x32 size tiles, 2 is double the tile size, and so on. 0
presents a map collapsed 10 times, or 320x320 pixel per tile.

Watch what happens to the frame rate as tile size increases.
"""

import sys

import pygame
from pygame import Color, K_ESCAPE, K_g, K_l, K_0, K_9

import paths
import gummworld2
from gummworld2 import data, model, geometry, toolkit
from gummworld2 import Engine, State, TiledMap, Vec2d


__version__ = '$Id: 09_collapse_map.py 437 2014-03-26 06:04:50Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


class Avatar(model.Object):
    
    def __init__(self, map_pos, screen_pos):
        model.Object.__init__(self)
        self.image = pygame.Surface((10, 10))
        self.rect = self.image.get_rect()
        pygame.draw.circle(self.image, Color('yellow'), self.rect.center, 4)
        self.image.set_colorkey(Color('black'))
        self.position = map_pos
        self.screen_position = screen_pos


class App(Engine):
    
    def __init__(self, resolution=(800, 600)):
        
        resolution = Vec2d(resolution)
        self.map_file_name = data.filepath('map', 'Gumm no swamps.tmx')
        try:
            tiled_map = TiledMap(self.map_file_name)
        except pygame.error:
            print('Sorry! Gummworld2 cannot include The Mana World assets' +
                  ' due to licensing.')
            print('To use this demo please download gummworld2_tmw.zip and' +
                  ' follow its instructions.')
            sys.exit()
        
        # Caption for window, and HUD in full-screen mode
        caption = '09 Collapse Map - G: grid | L: labels | Collapse: 1-10 (0 is 10)'
        
        Engine.__init__(self, caption=caption, resolution=resolution, camera_target=Avatar((325, 420), resolution // 2),
                        map=tiled_map, frame_speed=0)
        
        self.visible_objects = []
        
        # The collapse stat for the hud.
        self.collapse = 1
        
        # I like huds.
        def _get_collapse():
            return self.collapse

        def _get_tile_size():
            return State.map.tile_width, State.map.tile_height

        def map_info():
            layern = len(State.map.layers)
            tilen = 0
            for layer in State.map.layers:
                tilen += len(layer)
            return layern, tilen
        self.hud = toolkit.make_default_hud(caption)
        self.hud.add('Collapse', 'Collapse {}', 0, callback=_get_collapse)
        self.hud.add('Tile size', 'Tile size {}x{}', 0, 0, callback=_get_tile_size)
        self.hud.add('Layers/Tiles', 'Layers/Tiles {}/{}', 0, 0, callback=map_info)
        
        # Create a speed box for converting mouse position to destination
        # and scroll speed.
        self.speed_box = geometry.Diamond(0, 0, 4, 2)
        self.speed_box.center = State.camera.abs_screen_center
        self.max_speed_box = float(self.speed_box.width) / 2.0
        
        # Mouse and movement state. move_to is in world coordinates.
        self.move_to = None
        self.speed = None
        self.mouse_down = False
        
        self.grid_cache = {}
        self.label_cache = {}
        
    def update(self, dt):
        """overrides Engine.update"""
        # If mouse button is held down update for continuous walking.
        if self.mouse_down:
            self.update_mouse_movement(pygame.mouse.get_pos())
        self.update_camera_position()
        State.camera.update()
        self.visible_objects = toolkit.get_object_array()
        self.hud.update()
        
    def update_mouse_movement(self, pos):
        # Angle of movement.
        angle = geometry.angle_of(self.speed_box.center, pos)
        # Final destination.
        self.move_to = None
        for edge in self.speed_box.edges:
            # line_intersects_line() returns False or (True,(x,y)).
            cross = geometry.line_intersects_line(edge, (self.speed_box.center, pos))
            if cross:
                x, y = cross[0]
                self.move_to = State.camera.screen_to_world(pos)
                self.speed = geometry.distance(self.speed_box.center, (x, y)) / self.max_speed_box
                break
        
    def update_camera_position(self):
        """update the camera's position if any movement keys are held down
        """
        if self.move_to is not None:
            camera = State.camera
            wx, wy = camera.position
            # Speed formula.
            speed = self.speed * State.speed
            # If we're within spitting distance, then taking a full step would
            # overshoot the desired destination. Therefore, we'll jump there.
            if geometry.distance((wx, wy), self.move_to) < speed:
                wx, wy = self.move_to
                self.move_to = None
            else:
                # Otherwise, calculate the full step.
                angle = geometry.angle_of((wx, wy), self.move_to)
                wx, wy = geometry.point_on_circumference((wx, wy), speed, angle)
            # Keep avatar inside map bounds.
            rect = State.world.rect
            wx = max(min(wx, rect.right), rect.left)
            wy = max(min(wy, rect.bottom), rect.top)
            camera.position = wx, wy
        
    def draw(self, interp):
        """overrides Engine.draw"""
        # Draw stuff.
        State.screen.clear()
        toolkit.draw_object_array(self.visible_objects)
        if State.show_grid:
            toolkit.draw_grid(self.grid_cache)
        if State.show_labels:
            toolkit.draw_labels(self.label_cache)
        self.hud.draw(State.screen.surface)
        self.draw_avatar()
        if State.name == 'small':
            pygame.draw.rect(State.screen.surface, (99, 99, 99), State.camera.view.parent_rect, 1)
        State.screen.flip()
        
    def draw_avatar(self):
        camera = State.camera
        avatar = camera.target
        camera.surface.blit(avatar.image, camera.screen_center)
        
    def on_mouse_button_down(self, pos, button):
        self.mouse_down = True
        
    def on_mouse_button_up(self, pos, button):
        self.mouse_down = False
        
    def on_key_down(self, unicode, key, mod):
        # Turn on key-presses.
        if key == K_g:
            State.show_grid = not State.show_grid
        elif key == K_l:
            State.show_labels = not State.show_labels
        elif K_0 <= key <= K_9 + 1:
            if key == K_0:
                n = 10
            else:
                n = key - K_0
            if n == self.collapse:
                return
            self.map = TiledMap(self.map_file_name, (n, n))
            State.world.rect = self.map.rect.copy()
            State.map = self.map
            self.collapse = n
            self.grid_cache.clear()
        elif key == K_ESCAPE:
            self.pop()
        
    def on_quit(self):
        self.pop()


if __name__ == '__main__':
    app = App()
    gummworld2.run(app)
