#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""23_supermap.py - An example of using SuperMap in Gummworld2."""

import sys
import cProfile
import pstats

import pygame
from pygame import K_UP, K_DOWN, K_LEFT, K_RIGHT, K_ESCAPE, K_0

import paths
import gummworld2
from gummworld2 import Engine, State, SuperMap, MapHandler, TiledMap
from gummworld2 import data, toolkit, model, supermap


__version__ = '$Id: 23_supermap.py 437 2014-03-26 06:04:50Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


COLLAPSE_KEYS = range(K_1, K_9 + 1)


class App(Engine):
    
    def __init__(self):
        # resolution = 320, 240
        resolution = 640,480
        # resolution = 800,600
        self.movex = 0
        self.movey = 0
        self.visible_tiles = {}
        
        # Init Engine, no map or world.
        Engine.__init__(self, resolution=resolution, camera_target=model.Object(), frame_speed=0)

        # Create the supermap and world.
        make_supermap(self)
        self.camera.init_position(self.map.current.rect.center)
        
        # Show me da goods.
        def get_map_name():
            return State.map.current.name

        def get_num_visible():
            return len(State.map.visible_maps)

        def get_history():
            return len(State.map.history), State.map.max_maps

        def get_collapse():
            return State.map.current.collapse_level
        theme = toolkit.HUD_THEMES['vera']
        theme.update(dict(color='skyblue1'), gcolor='skyblue3', ocolor='black', scolor=None)
        self.hud = toolkit.make_default_hud(**theme)
        self.hud.add('Current', 'Current Map {}', get_map_name(), callback=get_map_name)
        self.hud.add('Visible', 'Visible {}', 0, callback=get_num_visible)
        self.hud.add('History', 'History {}/{}', 0, 0, callback=get_history)
        self.hud.add('Collapse Level', 'Collapse {}', 0, callback=get_collapse)
        self.clock.schedule_interval(self.hud.update, 0.2)
        
        State.speed = 3.33
    
    def update(self, dt):
        if self.movex or self.movey:
            r = State.camera.rect.move(self.movex, self.movey)
            r.clamp_ip(State.map.rect)
            State.camera.position = r.center
        State.camera.update(dt)
        State.map.update(dt)

    def draw(self, interp):
        State.screen.clear()
        State.map.draw()
        self.hud.draw(State.screen.surface)
        State.screen.flip()
    
    def on_key_down(self, unicode, key, mod):
        if key == K_DOWN:
            self.movey += State.speed
        elif key == K_UP:
            self.movey += -State.speed
        elif key == K_RIGHT:
            self.movex += State.speed
        elif key == K_LEFT:
            self.movex += -State.speed
        elif key in COLLAPSE_KEYS:
            clevel = key - K_0
            for mh in State.map.handlers.values():
                mh.collapse(clevel)
        elif key == K_ESCAPE:
            self.pop()
    
    def on_key_up(self, key, mod):
        if key == K_DOWN:
            self.movey -= State.speed
        elif key == K_UP:
            self.movey -= -State.speed
        elif key == K_RIGHT:
            self.movex -= State.speed
        elif key == K_LEFT:
            self.movex -= -State.speed
    
    def on_quit(self):
        self.pop()


class TiledMapHandler(MapHandler):
    
    # 4 gives better FPS at resolution 320x320, but can cause loading hiccups
    # 2 gives smoothest loading, highest cache hits
    collapse_level = (4, 4)
    
    def __init__(self, name, map_file_name):
        super(TiledMapHandler, self).__init__(name)
        self.map_file_name = map_file_name
    
    def load(self):
        try:
            self.map = TiledMap(
                data.filepath('map', self.map_file_name),
                collapse=self.collapse_level)
        except pygame.error:
            print('Sorry! Gummworld2 cannot include The Mana World assets' +
                  ' due to licensing.')
            print('To use this demo please download gummworld2_tmw.zip and' +
                  ' follow its instructions.')
            sys.exit()
    
    def collapse(self, clevel):
        if clevel > (1, 1):
            self.collapse_level = clevel
            # self.map could be None if the map isn't loaded or has been
            # unloaded.
            if self.map:
                self.map.collapse(clevel)


def make_supermap(app):
    # Make a 9x9 supermap, using the same map file for each.
    #map_filename = 'Gumm multi layer.tmx'
    map_filename = 'Gumm super desert.tmx'
    app.map = State.map = SuperMap()
    for n in ((0, 0),) + supermap.NEIGHBORS:
        app.map.add_handler(TiledMapHandler(n, map_filename))
    app.world = State.world = model.NoWorld(app.map.rect)


def main():
    app = App()
    gummworld2.run(app)


if __name__ == '__main__':
    if False:
        cProfile.run('main()', 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('time').print_stats()
    else:
        main()
