#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""25_spatialhash_geometry_test.py - A stress test and sandbox for advanced
geometrec objects in Gummworld2.
"""

import sys
from random import randrange, choice
import cProfile
import pstats

if sys.version_info[0] == 3:
    # This is dirty but range can be expensive with large sequences in
    # Python 2.
    xrange = range

import pygame
from pygame import Color, K_PLUS, K_MINUS, K_EQUALS, K_ESCAPE, K_g

import paths
import gummworld2
from gummworld2 import State, Engine, Vec2d, HUD
from gummworld2 import geometry, spatialhash, toolkit


__version__ = '$Id: 25_spatialhash_geometry_test.py 422 2013-08-21 04:30:47Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


class GeomBehavior(object):
    """Superclass providing movement, behavior, and drawing common to all
    entities.
    """
    
    size = 11, 11
    
    def __init__(self):
        choices = [-0.5, -0.3, -0.1, 0.1, 0.3, 0.5]
        self.step = Vec2d(choice(choices), choice(choices))
        self.hit = 0

    def update(self, *args):
        self.move()
        if self.hit:
            self.hit = 0
            self.image = self.image_red
        else:
            self.image = self.image_green
    
    def move(self):
        sr = self.rect
        wr = State.map.rect
        step = self.step
        self.position += step
        if not wr.contains(sr):
            if sr.x < wr.x or sr.right > wr.right:
                self.position -= (step.x, 0)
                step.x = -step.x
            if sr.y < wr.y or sr.bottom > wr.bottom:
                self.position -= (0, step.y)
                step.y = -step.y


class RectGeom(geometry.RectGeometry, GeomBehavior):
    """A rect with graphics, movement, collision detection, and behavior.
    """
    def __init__(self, position):
        geometry.RectGeometry.__init__(self, 0, 0, *self.size)
        GeomBehavior.__init__(self)
        
        self.image_green = pygame.surface.Surface(self.rect.size)
        self.image_green.fill(Color('green'))
    
        self.image_red = pygame.surface.Surface(self.rect.size)
        self.image_red.fill(Color('red'))
        
        self.image = self.image_green
        
        self.position = position


class CircleGeom(geometry.CircleGeometry, GeomBehavior):
    """A circle with graphics, movement, collision detection, and behavior.
    """
    def __init__(self, position):
        geometry.CircleGeometry.__init__(self, (0, 0), self.size[0] // 2)
        GeomBehavior.__init__(self)
        
        self.image_green = pygame.surface.Surface(self.rect.size)
        rect = self.image_green.get_rect()
        self.image_green.fill(Color('black'))
        self.image_green.set_colorkey(Color('black'))
        pygame.draw.circle(self.image_green, Color('green'), rect.center, self.radius)
        
        self.image_red = pygame.surface.Surface(self.rect.size)
        self.image_red.fill(Color('black'))
        self.image_red.set_colorkey(Color('black'))
        pygame.draw.circle(self.image_red, Color('red'), rect.center, self.radius)
        
        self.image = self.image_green
        
        self.position = position
    

class TriangleGeom(geometry.PolyGeometry, GeomBehavior):
    """A triangle with graphics, movement, collision detection, and behavior.
    """
    shape = (5, 0), (10, 10), (0, 10)
    
    def __init__(self, position):
        geometry.PolyGeometry.__init__(self, self.shape, position)
        GeomBehavior.__init__(self)
        
        self.image_green = pygame.surface.Surface(self.rect.size)
        self.image_green.fill(Color('black'))
        self.image_green.set_colorkey(Color('black'))
        pygame.draw.polygon(self.image_green, Color('green'), self.shape)
        
        self.image_red = pygame.surface.Surface(self.rect.size)
        self.image_red.fill(Color('black'))
        self.image_red.set_colorkey(Color('black'))
        pygame.draw.polygon(self.image_red, Color('red'), self.shape)
        
        self.image = self.image_green


class App(Engine):
    
    def __init__(self):

        self.tile_size = 60, 60
        self.map_size = 10, 10
        self.num_cells = 1
        self.worst_case = 0
        self.num_sprites = (
            200,    # slower misc
            500,    # Intel i3
            1000,   # Intel i5 or i7
            1500,   # D~Wave
        )[2]        # <- set this index appropriately for your CPU
        
        Engine.__init__(self, caption='24 SpatialHash Geometry Test - [+/-]: Cells | G: Grid', resolution=(600, 600),
                        tile_size=self.tile_size, map_size=self.map_size, update_speed=30, frame_speed=0)
        State.camera.init_position((300, 300))
        State.screen.fill_color = Color(33, 33, 0)
        
        # Make starting set of things.
        self.things = []
        world_rect = State.world.rect
        sprites_per_axis = int(self.num_sprites ** 0.5)
        remainder = self.num_sprites - sprites_per_axis * sprites_per_axis
        sizex, sizey = world_rect.w / sprites_per_axis, world_rect.h / sprites_per_axis
        for x in xrange(sprites_per_axis):
            for y in xrange(sprites_per_axis):
                self.things.append(self.make_thing((x * sizex + 5, y * sizey + 5)))
        for i in range(remainder):
            x = randrange(self.screen.size.x)
            y = randrange(self.screen.size.y)
            self.things.append(self.make_thing((x, y)))
        self.mouse_thing = RectGeom((300, 300))
        self.make_space()
        
        self.show_grid = True
        self.things_on_screen = 0, 0
        self.draw_level = 2
        
        self.num_collisions = 0
        self.collision_caret = 0
        self.collision_history = [100]
        self.fps_history = [100]
        self.collision_color = Color('red')
        self.minimum_ups_color = Color('blue')
        self.fps_color = Color('yellow')

        self.hud = None
        self.make_hud()
        State.show_hud = True
        self.clock.schedule_interval(self.hud.update, 0.2)

    @staticmethod
    def make_thing(pos):
        geom_class = choice([
            RectGeom,
            CircleGeom,
            TriangleGeom,
        ])
        return geom_class(pos)
    
    def make_space(self):
        cell_size = State.map.rect.w // self.num_cells
        State.world = self.world = spatialhash.SpatialHash(cell_size)
        State.world.addlist(self.things)
    
    def make_hud(self):
        def get_fps():
            return State.clock.fps

        def get_collisions():
            return self.num_collisions

        def get_things():
            return len(State.world)

        def get_cell_size():
            n = self.world.cell_size
            return n, n

        def get_collision_tests():
            num_things = len(State.world)
            return State.world.coll_tests / 1024.0, num_things * num_things / 1024.0

        self.hud = HUD(**toolkit.HUD_THEMES['roboto'])
        self.hud.set_fontsize(23)
        self.hud.set_color('tan')
        self.hud.set_gcolor('white')
        self.hud.set_ocolor('black')
        self.hud.set_owidth(0.5)

        # Frames per second.
        self.hud.add('FPS', 'FPS {:.0f}', get_fps(), callback=get_fps)
        
        # Collisions.
        self.hud.add('Collisions', 'Collisions {}', get_collisions(), callback=get_collisions)
        
        # Entities on screen.
        self.hud.add('Things', 'Things {}', get_things(), callback=get_things)
        
        # Num cells.
        self.hud.add('Num cells', 'Num cells {}x{}', *get_cell_size(), callback=get_cell_size)
        
        # Collision tests per tick / Node visits per tick.
        self.hud.add('Collision Tests', 'Collision Tests {:0.0f}K / {:0.0f}K', *get_collision_tests(),
                      callback=get_collision_tests)

        self.hud.add('Legend Collisions', 'Avg Collisions: Red line')
        self.hud.add('Legend FPS', 'Avg FPS: Yellow line')
        self.hud.add('Legend UPS', 'UPS {}: Blue line'.format(State.clock.max_ups))

    def update(self, dt):
        self.update_world()
        self.update_collisions()

    def update_world(self):
        world = State.world
        things = self.things
        for thing in things:
            thing.update()
        world.clear()
        world.addlist(things)
    
    def update_collisions(self):
        world = State.world
        world_rect = State.map.rect
        collisions = world.collidealldict()
        n = 0
        for e, c in collisions.items():
            e.hit = True
            n += len(c)
            for o in c:
                o.hit = True
        self.num_collisions = n

        y = 200 - self.num_collisions // 10
        history = self.collision_history
        history.append(y)
        if len(history) >= world_rect.width:
            history.pop(0)
        
        y = 200 - State.clock.fps // 5
        history = self.fps_history
        history.append(y)
        if len(history) >= world_rect.width:
            history.pop(0)
    
    def draw(self, interp):
        State.screen.clear()
        self.draw_world()
        self.draw_collisions()
        self.draw_grid()
        self.hud.draw(State.screen.surface)
        State.screen.flip()
    
    def draw_world(self):
        camera = State.camera
        blit = camera.surface.blit
        for thing in State.world.entities:
            blit(thing.image, thing.rect)
    
    def draw_collisions(self):
        camera = State.camera
        screen = camera.surface
        
        y = 200 - State.clock.max_ups // 5
        x1 = 0
        x2 = camera.rect.width
        pygame.draw.line(screen, self.minimum_ups_color, (x1, y), (x2, y))
        
        history = self.collision_history
        pygame.draw.lines(screen, self.collision_color, 0, tuple(zip(xrange(len(history)), history)))
        
        history = self.fps_history
        pygame.draw.lines(screen, self.fps_color, 0, tuple(zip(xrange(len(history)), history)))
    
    def draw_grid(self):
        if not self.show_grid:
            return
        if self.num_cells == max(*State.map.rect.size):
            return
        world = State.world
        wr = State.map.rect
        draw_line = pygame.draw.line
        surf = State.camera.view.surface
        color = Color('darkgrey')
        left, right, top, bottom = wr.left, wr.right, wr.top, wr.bottom
        minx = -1
        miny = -1
        cell_size = world.cell_size
        y_range = range(wr.y, wr.bottom + cell_size, cell_size)
        for x in range(wr.x, wr.right + cell_size, cell_size):
            for y in y_range:
                if x > minx:
                    minx = x
                    p1 = x, top
                    p2 = x, bottom
                    draw_line(surf, color, p1, p2)
                if y > miny:
                    miny = y
                    p1 = left, y
                    p2 = right, y
                    draw_line(surf, color, p1, p2)
    
    def on_key_down(self, unicode, key, mod):
        if key == K_g:
            # Toggle grid.
            self.show_grid = not self.show_grid
        elif key == K_ESCAPE:
            self.pop()
        elif key in (K_PLUS, K_EQUALS):
            # Resize world using smaller cell size. Minimum size is
            # max(*Thing.size).
            self.num_cells += 1
            cell_size = State.map.rect.w // self.num_cells
            if cell_size < max(*GeomBehavior.size):
                self.num_cells -= 1
            self.make_space()
            for thing in self.things:
                State.world.add(thing)
        elif key == K_MINUS:
            # Resize world using larger cell size. Maximum size is
            # max(*State.world.rect.size).
            if self.num_cells > 1:
                self.num_cells -= 1
                self.make_space()
                for thing in self.things:
                    State.world.add(thing)

    def on_quit(self):
        self.pop()

    def on_mouse_motion(self, pos, rel, buttons):
        self.mouse_thing.position = State.camera.screen_to_world(pos)
        State.world.add(self.mouse_thing)


if __name__ == '__main__':
    
    def run_it():
        gummworld2.run(app)
    
    app = App()

    if True:
        run_it()
    else:
        cProfile.run('run_it()')
        p = pstats.Stats()
        p.print_stats()
