#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""28_spatialhash_lines_test.py - A lines collision demo in Gummworld2."""

from random import randrange
import sys

if sys.version_info[0] == 3:
    # This is dirty but range can be expensive with large sequences in
    # Python 2.
    xrange = range


import pygame
from pygame import Color, K_ESCAPE, K_SPACE

import paths
import gummworld2
from gummworld2 import toolkit, State, SpatialHash, HUD, Vec2d
from gummworld2.geometry import LineGeometry, RectGeometry, CircleGeometry, PolyGeometry


__version__ = '$Id: 28_spatialhash_lines_test.py 407 2013-08-12 15:11:30Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


class App(gummworld2.Engine):
    
    def __init__(self):
        
        self.resolution = Vec2d(800, 600)
        camera_view_rect = pygame.Rect((0, 0), self.resolution)
        camera_target = gummworld2.model.Object(self.resolution / 2)
        
        gummworld2.Engine.__init__(self, caption='SpatialHash Lines - Space: Rect Cursor', resolution=self.resolution,
                                   camera_target=camera_target, camera_view_rect=camera_view_rect, update_speed=30,
                                   frame_speed=0)
        
        State.camera.init_position(self.resolution / 2)
        
        # Spatial hash for collision detection.
        cell_size = 40
        self.world = SpatialHash(cell_size)
        
        # Make line objects that can be drawn. Stick them in a spatial hash to
        # test collision with the mouse cursor.
        width, height = self.screen.rect.size
        for i in xrange(250):
            x1 = randrange(width)
            y1 = randrange(height)
            x2 = randrange(width)
            y2 = randrange(height)
            line = gummworld2.geometry.LineGeometry(x1, y1, x2, y2)
            line.color = Color('mediumpurple4')
            line.blend = 1
            self.collision_color = Color('orange')
            self.world.add(line)
        self.num_objects = len(self.world)

        self.collisions = []

        # Make some cursor shapes.
        self.cursor_color = Color('green')
        self.cursor_line = gummworld2.geometry.LineGeometry(0, 0, 0, 30)
        self.cursor_rect = gummworld2.geometry.RectGeometry(0, 0, 30, 30)
        self.cursor_circle = gummworld2.geometry.CircleGeometry((0, 0), 15)
        self.cursor_poly = gummworld2.geometry.PolyGeometry(((0, 0), (30, 0), (30, 30), (0, 30)))
        self.cursor = self.cursor_rect
        
        pygame.mouse.set_visible(0)

        # Make a HUD.
        def get_fps():
            return State.clock.fps

        def get_line_count():
            return self.num_objects

        def get_collisions():
            return len(self.collisions)
        theme = toolkit.HUD_THEMES['cherrycreamsoda'].copy()
        theme.update(dict(fontsize=32))
        self.hud = HUD(**theme)
        self.hud.add('FPS', 'FPS {:.0f}', get_fps(), callback=get_fps)
        self.hud.add('Lines', 'Lines {}', get_line_count(), callback=get_line_count)
        self.hud.add('Collisions', 'Collisions {}', get_collisions(), callback=get_collisions)
        self.clock.schedule_interval(self.hud.update, 0.1)

    def update(self, dt):
        self.collisions = self.world.collide(self.cursor)
    
    def draw(self, interp):
        screen = State.screen
        screen_surface = screen.surface
        draw_line = pygame.draw.line
        collisions = self.collisions
        screen.clear()
        for obj in self.world.entities:
            try:
                screen_surface.blit(obj.image, obj.rect)
            except AttributeError:
                if obj in collisions:
                    color = self.collision_color
                else:
                    color = obj.color
                draw_line(screen_surface, color, obj.p1, obj.p2, obj.blend)
        self.draw_cursor(screen_surface)
        self.hud.draw(screen_surface)
        screen.flip()
    
    def draw_cursor(self, surface):
        draw = pygame.draw
        cursor = self.cursor
        color = self.cursor_color
        if isinstance(self.cursor, LineGeometry):
            draw.aaline(surface, color, cursor.p1, cursor.p2, 1)
        elif isinstance(self.cursor, RectGeometry):
            draw.rect(surface, color, self.cursor.rect, 1)
        elif isinstance(self.cursor, CircleGeometry):
            draw.circle(surface, color, cursor.origin, cursor.radius, 1)
        elif isinstance(self.cursor, PolyGeometry):
            draw.polygon(surface, color, cursor.points, 1)
    
    def on_key_down(self, unicode, key, mod):
        if key == K_ESCAPE:
            self.pop()
        elif key == K_SPACE:
            if isinstance(self.cursor, gummworld2.geometry.LineGeometry):
                self.cursor = self.cursor_rect
                pygame.display.set_caption('SpatialHash Lines - Space: Rect Cursor')
            elif isinstance(self.cursor, gummworld2.geometry.RectGeometry):
                self.cursor = self.cursor_circle
                pygame.display.set_caption('SpatialHash Lines - Space: Circle Cursor')
            elif isinstance(self.cursor, gummworld2.geometry.CircleGeometry):
                self.cursor = self.cursor_poly
                pygame.display.set_caption('SpatialHash Lines - Space: Poly Cursor')
            elif isinstance(self.cursor, gummworld2.geometry.PolyGeometry):
                self.cursor = self.cursor_line
                pygame.display.set_caption('SpatialHash Lines - Space: Line Cursor')
            self.cursor.position = pygame.mouse.get_pos()
    
    def on_mouse_button_down(self, pos, button):
        pass
    
    def on_mouse_button_up(self, pos, button):
        pass
    
    def on_mouse_motion(self, pos, rel, buttons):
        self.cursor.position = pos
    
    def on_quit(self):
        self.pop()


if __name__ == '__main__':
    app = App()
    gummworld2.run(app)
