#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""29_basic_map_renderer.py - An example using the BasicMapRenderer in Gummworld2.

The BasicMapRenderer provides a very easy way to scroll the background tiles. It is as easy to use as the toolkit, and
in most cases requires far less memory than collapsing a map, without sacrificing high performance.

This demo produces a map of many 8x8 pixel tiles. Brute force rendering of this many tiny tiles would kill performance.
The renderer provides an intermediate rendering surface that can be drawn once and reused until it is marked dirty or
automatically retired when it is no longer being used.
"""

import pygame
from pygame import K_UP, K_DOWN, K_LEFT, K_RIGHT, K_ESCAPE

import paths
import gummworld2
from gummworld2 import State, Engine, BasicMapRenderer, Vec2d
from gummworld2 import toolkit


__version__ = '$Id: 29_basic_map_renderer.py 422 2013-08-21 04:30:47Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


class CameraTarget(object):
    
    def __init__(self, position=(0, 0)):
        self._position = Vec2d(position)
    
    # A camera target just needs to have a Vec2d position attribute. It helps
    # to protect it from overwriting with other types.
    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        self._position[:] = val


class App(Engine):
    
    def __init__(self):
        
        screen_size = Vec2d(800, 600)
        tile_size = 8, 8     # at 800x600 res: nearly 8,000 tiles on screen
        map_size = 300, 300  # at 300x300: 90,000 tiles in map
        
        self.movex = 0
        self.movey = 0
        
        Engine.__init__(self, resolution=screen_size, tile_size=tile_size, map_size=map_size,
                        camera_target=CameraTarget(screen_size // 2), frame_speed=0)
        
        self.scroll_speed = 10
        
        if __debug__:
            print('App: making tiles')
        toolkit.make_tiles(label=False)
        
        if __debug__:
            print('App: making renderer')
        self.renderer = BasicMapRenderer(State.map, max_scroll_speed=self.scroll_speed)
        
        if __debug__:
            print('App: making HUD')

        def count_map_tiles():
            basic_map = State.map
            ids = toolkit.get_visible_cell_ids(
                State.camera, basic_map, self.scroll_speed)
            layers = toolkit.get_objects_in_cell_ids(basic_map, ids)
            count = 0
            for layer in layers:
                count += len(layer)
            return count,

        def get_renderer_tiles():
            return len(self.renderer._tiles)
        self.hud = toolkit.make_default_hud()
        self.hud.add('BasicMap tiles: ', 'BasicMap tiles {}', 0, callback=count_map_tiles)
        self.hud.add('Renderer tiles: ', 'Renderer tiles {}', 0, callback=get_renderer_tiles)
        self.clock.schedule_interval(self.hud.update, 0.1)
        
        if __debug__:
            print('App: running')
    
    def update(self, dt):
        if self.movex or self.movey:
            State.camera.position += self.movex, self.movey
        State.camera.update()
        self.renderer.set_rect(center=State.camera.rect.center)

    def draw(self, interp):
        State.screen.clear()
        self.renderer.draw_tiles()
        self.hud.draw(State.screen.surface)
        State.screen.flip()

    def on_key_down(self, unicode, key, mod):
        if key == K_DOWN:
            self.movey += self.scroll_speed
        elif key == K_UP:
            self.movey += -self.scroll_speed
        elif key == K_RIGHT:
            self.movex += self.scroll_speed
        elif key == K_LEFT:
            self.movex += -self.scroll_speed
        elif key == K_ESCAPE:
            self.pop()
    
    def on_key_up(self, key, mod):
        if key == K_DOWN:
            self.movey -= self.scroll_speed
        elif key == K_UP:
            self.movey -= -self.scroll_speed
        elif key == K_RIGHT:
            self.movex -= self.scroll_speed
        elif key == K_LEFT:
            self.movex -= -self.scroll_speed
    
    def on_quit(self):
        self.pop()


if __name__ == '__main__':
    app = App()
    gummworld2.run(app)
