#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""30_renderer_dirty.py - An example of modifying the map in realtime with
the renderer in Gummworld2.
"""

import random
import sys

if sys.version_info[0] == 3:
    # This is dirty but range can be expensive with large sequences in
    # Python 2.
    xrange = range


import pygame
from pygame import Color, K_PLUS, K_MINUS, K_EQUALS, K_UP, K_DOWN, K_LEFT, K_RIGHT, K_ESCAPE

import paths
import gummworld2
from gummworld2 import State, Engine, BasicMapRenderer, Vec2d
from gummworld2 import toolkit


__version__ = '$Id: 30_renderer_dirty.py 422 2013-08-21 04:30:47Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


class CameraTarget(object):
    
    def __init__(self, position=(0, 0)):
        self._position = Vec2d(position)
    
    # A camera target just needs to have a Vec2d position attribute. It helps
    # to protect it from overwriting with other types.
    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        self._position[:] = val


class App(Engine):
    
    def __init__(self):
        
        screen_size = Vec2d(800, 600)
        tile_size = 32, 24
        map_size = 25, 25
        
        self.movex = 0
        self.movey = 0
        
        Engine.__init__(self, caption='30 Renderer Dirty - Num picks: +/-', resolution=screen_size,
                        tile_size=tile_size, map_size=map_size,
                        camera_target=CameraTarget(screen_size // 2),
                        frame_speed=0)
        
        self.scroll_speed = 4
        
        toolkit.make_tiles(label=True)
        self.fill_color = Color(50, 50, 50)
        self.frame_color = Color(0, 0, 0)
        self.dirty_color = Color(100, 50, 50)
        for tile in self.map.layers[0].objects.entities:
            self._color_image(tile.image, self.fill_color, self.frame_color)
        
        self.renderer = BasicMapRenderer(State.map, max_scroll_speed=self.scroll_speed)
        self.num_picks = 1
        self.picks = []
        self._pick_tiles()

        def count_map_tiles():
            """count tiles in map for the HUD"""
            basic_map = State.map
            ids = toolkit.get_visible_cell_ids(
                State.camera, basic_map, self.scroll_speed)
            layers = toolkit.get_objects_in_cell_ids(basic_map, ids)
            count = 0
            for layer in layers:
                count += len(layer)
            return count

        def get_renderer_tiles():
            return len(self.renderer._tiles)

        def get_num_picks():
            return self.num_picks

        def get_dirty_rects():
            return len(self.renderer.dirty_rects)
        self.hud = toolkit.make_default_hud(theme='roboto', fontsize=25, alpha=0.65)
        self.hud.add('BasicMap tiles', 'BasicMap tiles {}', 0, callback=count_map_tiles)
        self.hud.add('Renderer tiles', 'Renderer tiles {}', 0, callback=get_renderer_tiles)
        self.hud.add('Num picks', 'Num picks {}', 0, callback=get_num_picks)
        self.hud.add('Dirty rects', 'Dirty rects {}', 0, callback=get_dirty_rects)
        
        self.clock.schedule_interval(self._pick_tiles, 2.0)
    
    def update(self, dt):
        if self.movex or self.movey:
            State.camera.position += self.movex, self.movey
        State.camera.update()
        self.renderer.set_rect(center=State.camera.rect.center)
        self._update_tiles()
        self.hud.update(dt)
    
    def _update_tiles(self):
        """modify tile picks"""
        self._up_dirty_color()
        color = self.dirty_color
        # modify the pixels of the background tiles that are picked
        for tile in self.picks:
            self._color_image(tile.image, color, self.frame_color)
        # set the renderer dirty
        self.renderer.set_dirty(*[t.rect for t in self.picks])

    def _up_dirty_color(self):
        """bump color up in brightness"""
        color = self.dirty_color
        if color.r < 253:
            color.r += 2
            color.g += 1
            color.b += 1
    
    @staticmethod
    def _color_image(image, fill_color, frame_color):
        """apply color to an image"""
        image.fill(fill_color)
        pygame.draw.rect(image, frame_color, image.get_rect(), 1)
    
    def _pick_tiles(self, *args):
        """pick random tiles to modify"""
        self.dirty_color.r = 100
        self.dirty_color.g = 50
        self.dirty_color.b = 50
        # restore the original color to the picked map tiles
        for tile in self.picks:
            self._color_image(tile.image, self.fill_color, self.frame_color)
        del self.picks[:]
        for i in range(self.num_picks):
            self.picks.append(random.choice(
                tuple(self.map.layers[0].objects.entities)))
    
    def _add_pick(self):
        """add a tile pick"""
        self.num_picks += 1
        self.picks.append(random.choice(
            tuple(self.map.layers[0].objects.objects)))
    
    def _remove_pick(self):
        """remove a tile pick"""
        if len(self.picks) > 1:
            self.num_picks -= 1
            tile = self.picks.pop(0)
            # restore the original color to the map tile, and set the renderer dirty
            self._color_image(tile.image, self.fill_color, self.frame_color)
            self.renderer.set_dirty(tile.rect)
    
    def draw(self, interp):
        State.screen.clear()
        self.renderer.draw_tiles()
        self.hud.draw(State.screen.surface)
        State.screen.flip()

    def on_key_down(self, unicode, key, mod):
        if key == K_DOWN:
            self.movey += self.scroll_speed
        elif key == K_UP:
            self.movey += -self.scroll_speed
        elif key == K_RIGHT:
            self.movex += self.scroll_speed
        elif key == K_LEFT:
            self.movex += -self.scroll_speed
        elif key in (K_EQUALS, K_PLUS):
            self._add_pick()
        elif key == K_MINUS:
            self._remove_pick()
        elif key == K_ESCAPE:
            self.pop()
    
    def on_key_up(self, key, mod):
        if key == K_DOWN:
            self.movey -= self.scroll_speed
        elif key == K_UP:
            self.movey -= -self.scroll_speed
        elif key == K_RIGHT:
            self.movex -= self.scroll_speed
        elif key == K_LEFT:
            self.movex -= -self.scroll_speed

    def on_quit(self):
        self.pop()


if __name__ == '__main__':
    app = App()
    gummworld2.run(app)
